/* main.cpp  - convert .mesh into postscript file */
/***************************************************************************
    begin                : 2003-03-06
    copyright            : (C) 2003 by Gunter Winkler
    email                : guwi17@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>

#include "cmdline.h"

template <class STREAM, class STRING>
void read_next_line(STREAM& ins, STRING& line)
{
  do {
	line.clear();
	if (ins.eof()) return;
	getline(ins, line);
  } while ( (line.length() == 0) || (line[0] == '#') );
}
/*
 * reads next word from stream
 * skips rest of line if word is "#" (ignore comments)
 * reads double quoted words like "a b c d" and "foo" as one word 
 * (and strips '"')
 */
template <class STREAM, class STRING>
void read_next_word(STREAM& ins, STRING& word)
{
  do {
	word = "";
	if (ins.eof()) return;
	ins >> word;
	if ((word.length()>0) && (word[0]=='#')) {
	  getline(ins, word);
	  continue;
	}
	if ((word.length()>0) && (word[0]=='"')) {
	  if (word[word.length()-1]='"') {
		word.erase(word.length()-1,1);
		word.erase(0,1);
	  } else {
		STRING    word2;
		getline(ins, word2, '"');
		word.erase(0,1);
		word += word2;
	  }
	}
  } while ( (word.length() == 0) );
}

using std::cin;
using std::cout;
using std::cerr;
using std::endl;
using std::flush;

int main(int argc, char** argv)
{
  gengetopt_args_info   options;
  std::string           current_line;
  std::string           current_word;
  std::string           word;
  std::stringstream     ins;

  const double  size = 500.0;

  int linecount=0;

  int dimension=0;
  int nv=0;         // number of vertices
  double  *vertex=0;  
  int ne=0;         // number of edges
  int     *edge=0;
  int nt=0, *tri=0;   // number of triangles, 2-by-nt list of node numbers
  int nq=0, *quad=0;  // number of quadrilaterals, 4-by-nq list of node numbers

  cmdline_parser(argc, argv, &options);

  std::streambuf   *inbuf, *outbuf;
  std::ifstream    ifs;
  std::ofstream    ofs;
  if (options.input_given) {
	ifs.open( options.input_arg );
	if (ifs.fail()) {
	  cerr << "cannot open input file '" << options.input_arg << "'" << endl;
	  return -1;
	}
	inbuf = ifs.rdbuf();
	cin.rdbuf(inbuf);
  }
  if (options.output_given) {
	ofs.open( options.output_arg );
	if (ofs.fail()) {
	  cerr << "cannot open otput file '" << options.output_arg << "'" << endl;
	  return -1;
	}
	outbuf = ofs.rdbuf();
	cout.rdbuf(outbuf);
  }

  while (!cin.eof()) {
	++linecount;
	read_next_word(cin, current_word);
	word = current_word;
	for(size_t i=0; i<word.length(); ++i) {
	  word[i] = toupper(word[i]);
	}
	if ("DIMENSION"==word) {
	  /* read: "Dimension", int dimension */
	  read_next_word(cin, current_word);
	  ins << current_word;
	  ins >> dimension;
	  ins.clear();
#ifndef NDEBUG
	  cerr << "dimension is " << dimension << endl;
#endif
	  continue;
	}
	if ("VERTICES"==word) {
	  if (0==dimension) {
		cerr << "error: dimension must be set before vertices" << endl;
		return 1;
	  }
	  /* read: "Vertices", int nv */
	  /* read: double x_1, ..., x_{nv}, int flag */
	  read_next_word(cin, current_word);
	  ins << current_word;
	  ins >> nv;
	  ins.clear();
#ifndef NDEBUG
	  cerr << "read " << nv << " vertices\n";
#endif
	  if (nv==0) {
		cerr << "error: number of vertices is zero" << endl;
		return 2;
	  }
	  vertex = new double[nv*dimension];
	  int j=0;
	  for (int i=0; i<nv; ++i) {
		double dtemp;
		for (int jj=0; jj<dimension; ++jj) {
		  read_next_word(cin, current_word);
		  ins << current_word;
		  ins >> dtemp;
		  vertex[j] = dtemp;
		  ins.clear();
		  ++j;
		};
		read_next_word(cin, current_word); // flag
	  }
	  continue;
	}
	if ("EDGES"==word) {
	  /* read: "Edges", int ne */
	  /* read: int v_1, v_2, int flag */
	  read_next_word(cin, current_word);
	  ins << current_word;
	  ins >> ne;
	  ins.clear();
#ifndef NDEBUG
	  cerr << "read " << ne << " edges\n";
#endif
	  if (ne==0) {
		cerr << "warning: number of edges is zero" << endl;
	  }
	  edge = new int[ne*2];
	  int j=0;
	  for (int i=0; i<ne; ++i) {
		int itemp;
		for (int jj=0; jj<2; ++jj) {
		  read_next_word(cin, current_word);
		  ins << current_word;
		  ins >> itemp;
		  edge[j] = itemp;
		  ins.clear();
		  ++j;
		};
		read_next_word(cin, current_word); // flag
	  }
	  continue;
	}
	if ("TRIANGLES"==word) {
	  /* read: "Triangles", int nt */
	  /* read: int v_1, v_2, v_3, int flag */
	  const int NN=3;
	  read_next_word(cin, current_word);
	  ins << current_word;
	  ins >> nt;
	  ins.clear();
#ifndef NDEBUG
	  cerr << "read " << nt << " Triangles\n";
#endif
	  if (nt==0) {
		cerr << "warning: number of triangles is zero" << endl;
	  }
	  tri = new int[nt*NN];
	  int j=0;
	  for (int i=0; i<nt; ++i) {
		int itemp;
		for (int jj=0; jj<NN; ++jj) {
		  read_next_word(cin, current_word);
		  ins << current_word;
		  ins >> itemp;
		  tri[j] = itemp;
		  ins.clear();
		  ++j;
		};
		read_next_word(cin, current_word); // flag
	  }
	  continue;
	}
	if ("QUADRILATERALS"==word) {
	  /* read: "Quadrilaterals", int nq */
	  /* read: int v_1, v_2, v_3, v_4, int flag */
	  read_next_word(cin, current_word);
	  ins << current_word;
	  ins >> nq;
	  ins.clear();
#ifndef NDEBUG
	  cerr << "read " << nq << " Quadrilaterals\n";
#endif
	  if (nq==0) {
		cerr << "warning: number of quadrilaterals is zero" << endl;
	  }
	  quad = new int[nq*4];
	  int j=0;
	  for (int i=0; i<nq; ++i) {
		int itemp;
		for (int jj=0; jj<4; ++jj) {
		  read_next_word(cin, current_word);
		  ins << current_word;
		  ins >> itemp;
		  quad[j] = itemp;
		  ins.clear();
		  ++j;
		};
		read_next_word(cin, current_word); // flag
	  }
	  continue;
	}
#ifndef NDEBUG
	cerr << "skipped: " << current_word << endl;
#endif
  }

  /* assume 2D */
  double xmin=vertex[0], xmax=xmin;
  double ymin=vertex[1], ymax=ymin;
  double xsc, ysc;
  for (int i=0; i<nv; ++i) {
	if (vertex[i*dimension+0]<xmin) xmin=vertex[i*dimension+0];
	if (vertex[i*dimension+1]<ymin) ymin=vertex[i*dimension+1];
	if (vertex[i*dimension+0]>xmax) xmax=vertex[i*dimension+0];
	if (vertex[i*dimension+1]>ymax) ymax=vertex[i*dimension+1];
  }

  xsc = size / (xmax-xmin);
  ysc = size / (ymax-ymin);

  cout <<
	"%!PS-Adobe-2.0 EPSF-2.0\n"
	"%%Title: ";
  if (options.input_given) {
	cout << options.input_arg;
  } else {
	cout << "(stdin)";
  }
  cout << "\n"
	"%%Creator: mesh2ps\n"
	"%%BoundingBox: -1 -1 " << size << " " << size << "\n"
	"%%Orientation: Portrait\n"
	"%%Magnification: 1.0000"
	"%%EndComments\n"
	;


  cout.setf(std::ios::fixed);
  cout.precision(1);

  cout << "gsave" << endl;

  cout << "% (boundary) edges\n";

  cout << "0.0 0.0 0.0 setrgbcolor\n";
  cout << "1.0 setlinewidth\n";
  for (int i=0; i<ne; ++i) {
	cout << "newpath\n";
	int e = edge[2*i]-1;
	cout << (xsc*(vertex[dimension*e]-xmin)) << " ";
	cout << (ysc*(vertex[dimension*e+1]-ymin)) << " ";
	cout << "moveto ";
	e = edge[2*i+1]-1;
	cout << (xsc*(vertex[dimension*e]-xmin)) << " ";
	cout << (ysc*(vertex[dimension*e+1]-ymin)) << " ";
	cout << "lineto\n";
	cout << "closepath stroke\n";
  }
  cout << flush;

  cout << "% triangles\n";
  cout << "0.5 0.0 0.0 setrgbcolor\n";
  cout << "0.1 setlinewidth\n";
  for (int i=0; i<nt; ++i) {
	const int NN = 3;
	cout << "newpath\n";
	int e = tri[NN*i+NN-1]-1;
	cout << (xsc*(vertex[dimension*e]-xmin)) << " ";
	cout << (ysc*(vertex[dimension*e+1]-ymin)) << " ";
	cout << "moveto ";
	for (int j=0; j<NN; ++j) {
	  e = tri[NN*i+j]-1;
	  cout << (xsc*(vertex[dimension*e]-xmin)) << " ";
	  cout << (ysc*(vertex[dimension*e+1]-ymin)) << " ";
	  cout << "lineto ";
	}
	cout << "\nclosepath stroke\n";
  }
  cout << flush;


  cout << "% quadrilaterals\n";
  cout << "0.0 0.0 0.5 setrgbcolor\n";
  cout << "0.1 setlinewidth\n";
  for (int i=0; i<nq; ++i) {
	cout << "newpath\n";
	int e = quad[4*i+3]-1;
	cout << (xsc*(vertex[dimension*e]-xmin)) << " ";
	cout << (ysc*(vertex[dimension*e+1]-ymin)) << " ";
	cout << "moveto ";
	for (int j=0; j<4; ++j) {
	  e = quad[4*i+j]-1;
	  cout << (xsc*(vertex[dimension*e]-xmin)) << " ";
	  cout << (ysc*(vertex[dimension*e+1]-ymin)) << " ";
	  cout << "lineto\n";
	}
	cout << "closepath stroke\n";
  }
  cout << flush;

  cout <<
	"grestore\n"
	;
  
  cout << flush;

  if (vertex)  delete[] vertex;
  if (edge)    delete[] edge;
  if (tri)     delete[] tri;
  if (quad)    delete[] quad;
  return 0;
}
